import unittest
from unittest.mock import patch
import pandas as pd
from rbceq2.core_logic.alleles import Allele, Line

# Updated imports:
from rbceq2.db.db import (
    Db,
    VariantCountMismatchError,
    compare_antigen_profiles,
    DbDataConsistencyChecker,  # New import
    prepare_db,  # New import
    # Keep existing antigen parsing helpers if they are directly tested or needed for test setup
    # For now, assuming they are not directly tested here, but used by compare_antigen_profiles
    # _NUM_ID_RE, _ALPHA_CANON_RE, _canonical_alpha, Antigen, AntigenParser, NumericParser, AlphaParser
)


class TestVariantCountMismatchError(unittest.TestCase):
    def test_error_message(self):
        err = VariantCountMismatchError("A,B", "X,Y,Z")
        self.assertIn("Number of GRCh37 variants must equal", str(err))
        self.assertIn("A,B", str(err))
        self.assertIn("X,Y,Z", str(err))


class TestDb(unittest.TestCase):
    def setUp(self):
        self.headers = "GRCh37\tGRCh38\tChrom\tGenotype\tPhenotype_change\tGenotype_alt\tPhenotype_alt_change\tLane\tSub_type\tWeight_of_genotype\tReference_genotype\tAntithetical\tPhenotype\tPhenotype_alt\n"

        self.good_csv_content = (
            self.headers
            + """1:100_G_A\t1:100_G_A\tchr1\tBG*01.01\tSYS:1w\tBG*01.01X\tSYS:X+weak\tTrue\tSubA\t100\tNo\tNo\tSYS:1w\tSYS:X+weak
1:200_T_C\t1:200_T_C\tchr1\tREF*01.02\tSYS:2\tREF*01.02X\tSYS:Y+\tFalse\tSubA\t50\tYes\tNo\tSYS:2\tSYS:Y+
1:300_T_C,1:301_A_G\t1:300_T_C,1:301_A_G\tchr1\tANTI*01.03\tSYS:3\tANTI*01.03X\tSYS:Z+\tTrue\tSubB\t1\tNo\tYes\tSYS:3\tSYS:Z+
2:400_A_G\t2:400_A_G\tchr2\tREF*02.01\t.\t.\t.\tFalse\tSubC\t1\tYes\tNo\t.\t.
"""
        )
        self.mismatch_grch_csv_content = (
            self.headers
            + """1:300_T_C,1:301_A_G\t1:300_T_C\tchr1\tBG*01.03\t.\t.\t.\tTrue\tSubB\t1\tNo\tYes\t.\t.
"""
        )
        # KEL:1 (numeric, positive) vs KEL:K- (alpha, negative) for Phenotype_change/Phenotype_alt_change
        # The Phenotype/Phenotype_alt are KEL:1 vs KEL:K+ (consistent) to not trigger mismatch there
        self.antigen_mismatch_csv_content = (
            self.headers
            + """1:100_G_A\t1:100_G_A\tchr1\tBG*01.01\tKEL:1\tBG*01.01X\tKEL:K-\tTrue\tSubA\t100\tNo\tNo\tKEL:1\tKEL:K+
"""
        )
        self.mock_load_db_path = "rbceq2.db.db.load_db"

        global ANTIGEN_MAP  # Use the global ANTIGEN_MAP defined later in the file
        if "SYS" not in ANTIGEN_MAP:
            ANTIGEN_MAP["SYS"] = {"1": "X", "2": "Y", "3": "Z"}
        if "KEL" not in ANTIGEN_MAP:
            ANTIGEN_MAP["KEL"] = {}
        ANTIGEN_MAP["KEL"]["1"] = "K"


    @patch("rbceq2.db.db.load_db")
    def test_db_creation_and_consistency_checks(self, mock_load_db_func):
        # Scenario 1: GRCh37/38 mismatch
        mock_load_db_func.return_value = self.mismatch_grch_csv_content
        df_mismatch = prepare_db()
        with self.assertRaises(VariantCountMismatchError):
            DbDataConsistencyChecker.run_all_checks(
                df=df_mismatch, ref_genome_name="GRCh37"
            )

        # Scenario 2: Antigen profile mismatch for Phenotype_change/Phenotype_alt_change
        mock_load_db_func.return_value = self.antigen_mismatch_csv_content
        df_antigen_mismatch = prepare_db()
        with self.assertRaises(
            AssertionError
        ):  # Expecting mismatch leading to AssertionError
            DbDataConsistencyChecker.run_all_checks(
                df=df_antigen_mismatch, ref_genome_name="GRCh37"
            )

        # Scenario 3: Good data
        mock_load_db_func.return_value = self.good_csv_content
        df_good = prepare_db()
        DbDataConsistencyChecker.run_all_checks(
            df=df_good, ref_genome_name="GRCh37"
        )  # Should pass
        db_obj = Db(ref="Genotype", df=df_good)

        self.assertIsInstance(db_obj.df, pd.DataFrame)
        self.assertEqual(db_obj.df.shape[0], 4)
        self.assertIn("chr1", db_obj.lane_variants)
        self.assertEqual(len(db_obj.lane_variants["chr1"]), 0)
        self.assertIn("REF", db_obj.reference_alleles)
        self.assertEqual(len(db_obj.reference_alleles), 1)
        self.assertEqual(db_obj.reference_alleles["REF"].genotype, "REF*02.01")
        self.assertIn("ANTI", db_obj.antitheticals)
        self.assertIn("ANTI*01.03", db_obj.antitheticals["ANTI"])

    def test_check_grch37_38_variant_counts_direct(self):
        df_mismatch = pd.DataFrame(
            {
                "GRCh37": ["1:300_T_C,1:301_A_G", "1:1_A_T"],
                "GRCh38": ["1:300_T_C", "1:1_A_T,1:2_C_G"],
            }
        )
        with self.assertRaises(VariantCountMismatchError) as cm:
            DbDataConsistencyChecker.check_grch37_38_variant_counts(df_mismatch)
        self.assertIn("1:300_T_C,1:301_A_G", str(cm.exception))
        self.assertIn("1:300_T_C", str(cm.exception))

        df_good = pd.DataFrame(
            {
                "GRCh37": ["1:300_T_C,1:301_A_G", "1:1_A_T"],
                "GRCh38": ["1:300_T_C,1:301_A_G", "1:1_A_T"],
            }
        )
        DbDataConsistencyChecker.check_grch37_38_variant_counts(df_good)

        df_empty_vs_nonempty = pd.DataFrame({"GRCh37": ["."], "GRCh38": ["1:1_A_T"]})
        with self.assertRaises(VariantCountMismatchError):
            DbDataConsistencyChecker.check_grch37_38_variant_counts(
                df_empty_vs_nonempty
            )

        df_both_empty = pd.DataFrame({"GRCh37": ["."], "GRCh38": ["."]})
        DbDataConsistencyChecker.check_grch37_38_variant_counts(df_both_empty)

    @patch("rbceq2.db.db.load_db")
    def test_line_generator_and_make_alleles(self, mock_load_db_func):
        mock_load_db_func.return_value = self.good_csv_content
        df = prepare_db()
        DbDataConsistencyChecker.run_all_checks(df=df, ref_genome_name="GRCh37")
        db_obj = Db(ref="Genotype", df=df)

        ref_df_subset = db_obj.df.query('Reference_genotype == "Yes"')
        lines = list(db_obj.line_generator(ref_df_subset))
        self.assertEqual(len(lines), 2)
        self.assertIsInstance(lines[0], Line)
        self.assertEqual(lines[0].geno, "REF*01.02")

        all_alleles = list(db_obj.make_alleles())
        self.assertEqual(len(all_alleles), 4)
        for a in all_alleles:
            self.assertIsInstance(a, Allele)

        allele_bg0101 = next(a for a in all_alleles if a.genotype == "BG*01.01")
        # The defining_variants construction adds chrom: before the variant string
        # and assumes GRCh37 column from self.ref being Genotype, and Line.allele_defining_variants
        # getting self.df[self.ref] as input, which is 'Genotype'.
        # This seems like a slight mix-up. `make_alleles` actually uses `line.allele_defining_variants`
        # which is populated from `df[self.ref]`. If `self.ref` is "Genotype", then `line.allele_defining_variants`
        # will be "BG*01.01", not "1:100_G_A".
        # The Line dataclass has `allele_defining_variants: str` which gets `df[self.ref]`.
        # This should be df['GRCh37'] or df['GRCh38'] depending on the desired reference for defining variants.
        # Let's assume `self.ref` is intended to be the column containing the defining variants like "1:100_G_A"
        # For `make_alleles` to work as intended (creating defining_variants like "chr1:1:100_G_A"),
        # `db_obj` should be initialized with `ref="GRCh37"` (or "GRCh38").
        # Let's re-initialize db_obj for this specific test part.

        db_obj_for_alleles = Db(
            ref="GRCh37", df=df
        )  # Use GRCh37 as the source for defining variants
        all_alleles_corrected = list(db_obj_for_alleles.make_alleles())
        allele_bg0101_corrected = next(
            a for a in all_alleles_corrected if a.genotype == "BG*01.01"
        )
        # line.allele_defining_variants will be "1:100_G_A" from df["GRCh37"]
        # and line.chrom will be "chr1"
        self.assertEqual(
            allele_bg0101_corrected.defining_variants, frozenset({"1:1:100_G_A"})
        )


# ANTIGEN_MAP: dict[str, dict[str, str]] = {
#     "ABCC1": {"1": "WLF"}, "ATP11C": {"1": "LIL"},
#     "AUG": {"1": "AUG1", "2": "At(a", "3": "ATML", "4": "ATAM"},
#     "CD59": {"1": "CD59.1"},
#     "CO": {"1": "Co(a", "2": "Co(b", "3": "Co3", "4": "Co4"},
#     "CROM": {"1": "Cr(a", "10": "UMC", "11": "GUTI", "12": "SERF", "13": "ZENA", "14": "CROV", "15": "CRAM", "16": "CROZ", "17": "CRUE", "18": "CRAG", "19": "CROK", "2": "Tc(a", "20": "CORS", "3": "Tc(b", "4": "Tc(c", "5": "Dr(a", "6": "Es(a", "7": "IFC", "8": "WES(a", "9": "WES(b"},
#     "CTL2": {"1": "VER", "2": "RIF", "3": "Cs(a", "4": "Cs(b", "5": "BROS"},
#     "DI": {"1": "Di(a", "10": "Bp(a", "11": "Mo(a", "12": "Hg(a", "13": "Vg(a", "14": "Sw(a", "15": "BOW", "16": "NFLD", "17": "Jn(a", "18": "KREP", "19": "Tr(a", "2": "Di(b", "20": "Fr(a", "21": "SW1", "22": "DISK", "23": "DIST", "3": "Wr(a", "4": "Wr(b", "5": "Wd(a", "6": "Rb(a", "7": "WARR", "8": "ELO", "9": "Wu"},
#     "DO": {"1": "Do(a", "10": "DODE", "2": "Do(b", "3": "Gy(a", "4": "Hy", "5": "Jo(a", "6": "DOYA", "7": "DOMR", "8": "DOLG", "9": "DOLC"},
#     "EMM": {"1": "Emm"},
#     "ER": {"1": "Er(a", "2": "Er(b", "3": "Er3", "4": "ERSA", "5": "ERAMA"},
#     "FORS": {"1": "FORS"},
#     "FY": {"1": "Fy(a", "2": "Fy(b", "3": "Fy3", "5": "Fy5", "6": "Fy6"},
#     "GE": {"10": "GEPL", "11": "GEAT", "12": "GETI", "13": "GECT", "14": "GEAR", "2": "Ge2", "3": "Ge3", "4": "Ge4", "5": "Wb", "6": "Ls(a", "7": "An(a", "8": "Dh(a", "9": "GEIS"},
#     "GIL": {"1": "GIL"}, "GLOB": {"1": "P"}, "I": {"1": "I"},
#     "IN": {"1": "In(a", "2": "In(b", "3": "INFI", "4": "INJA", "5": "INRA", "6": "INSL"},
#     "JK": {"1": "Jk(a", "2": "Jk(b", "3": "Jk3"},
#     "JMH": {"1": "JMH", "2": "JMHK", "3": "JMHL", "4": "JMHG", "5": "JMHM", "6": "JMHQ", "7": "JMHN", "8": "JMHA"},
#     "KANNO": {"1": "KANNO1"},
#     "KEL": {"1": "K", "10": "Ul(a", "11": "K11", "12": "K12", "13": "K13", "14": "K14", "16": "K16", "17": "K17", "18": "K18", "19": "K19", "2": "k", "20": "Km", "21": "Kp(c", "22": "K22", "23": "K23", "24": "K24", "25": "VLAN", "26": "TOU", "27": "RAZ", "28": "VONG", "29": "KALT", "3": "Kp(a", "30": "KTIM", "31": "KYO", "32": "KUCI", "33": "KANT", "34": "KASH", "35": "KELP", "36": "KETI", "37": "KHUL", "38": "KYOR", "39": "KEAL", "4": "Kp(b", "40": "KHIZ", "41": "KHOZ", "5": "Ku", "6": "Js(a", "7": "Js(b"},
#     "KN": {"1": "Kn(a", "10": "KDAS", "11": "DACY", "12": "YCAD", "13": "KNMB", "2": "Kn(b", "3": "McC(a", "4": "Sl1", "5": "Yk(a", "6": "McC(b", "7": "Vil", "8": "Sl3", "9": "KCAM"},
#     "LU": {"1": "Lu(a", "11": "Lu11", "12": "Lu12", "13": "Lu13", "14": "Lu14", "16": "Lu16", "17": "Lu17", "18": "Au(a", "19": "Au(b", "2": "Lu(b", "20": "Lu20", "21": "Lu21", "22": "LURC", "23": "LUIT", "24": "LUGA", "25": "LUAC", "26": "LUBI", "27": "LUYA", "28": "LUNU", "29": "LURA", "3": "Lu3", "30": "LUOM", "4": "Lu4", "5": "Lu5", "6": "Lu6", "7": "Lu7", "8": "Lu8", "9": "Lu9"},
#     "LW": {"5": "LWa", "6": "LWab", "7": "LWb", "8": "LWEM"},
#     "MAM": {"1": "MAM"},
#     "MNS": {"1": "M", "10": "Mur", "11": "Mg", "12": "Vr", "13": "Me", "14": "Mt(a", "15": "St(a", "16": "Ri(a", "17": "Cl(a", "18": "Ny(a", "19": "Hut", "2": "N", "20": "Hil", "21": "Mv", "22": "Far", "23": "sD", "24": "Mit", "25": "Dantu", "26": "Hop", "27": "Nob", "28": "En(a", "29": "ENKT", "3": "S", "30": "`N'", "31": "Or", "32": "DANE", "33": "TSEN", "34": "MINY", "35": "MUT", "36": "SAT", "37": "ERIK", "38": "Os(a", "39": "ENEP", "4": "s", "40": "ENEH", "41": "HAG", "42": "ENAV", "43": "MARS", "44": "ENDA", "45": "ENEV", "46": "MNTD", "47": "SARA", "48": "KIPP", "49": "JENU", "5": "U", "50": "SUMI", "6": "He", "7": "Mia", "8": "Mc", "9": "Vw"},
#     "OK": {"1": "Ok(a", "2": "OKGV", "3": "OKVM"},
#     "PEL": {"1": "PEL"}, "RAPH": {"1": "MER2"},
#     "RH": {"10": "V", "11": "Ew", "12": "G", "17": "Hro", "18": "Hr", "19": "hrS", "2": "C", "20": "VS", "21": "CG", "22": "CE", "23": "Dw", "26": "c_like", "27": "cE", "28": "hrH", "29": "Rh29", "3": "E", "30": "Goa", "31": "hrB", "32": "Rh32", "33": "Rh33", "34": "HrB", "35": "Rh35", "36": "Bea", "37": "Evans", "39": "Rh39", "4": "c", "40": "Tar", "41": "Rh41", "42": "Rh42", "43": "Crawford", "44": "Nou", "45": "Riv", "46": "Sec", "47": "Dav", "48": "JAL", "49": "STEM", "5": "e", "50": "FPTT", "51": "MAR", "52": "BARC", "53": "JAHK", "54": "DAK", "55": "LOCR", "56": "CENR", "57": "CEST", "58": "CELO", "59": "CEAG", "6": "f", "60": "PARG", "61": "CEVF", "62": "CEWA", "63": "CETW", "7": "Ce", "8": "Cw", "9": "Cx"},
#     "RHAG": {"1": "Duclos", "2": "Ol(a", "3": "DSLK", "5": "Kg", "6": "SHER", "7": "THIN"},
#     "RHD": {"1": "D"},
#     "SC": {"1": "Sc1", "2": "Sc2", "3": "Sc3", "4": "Rd", "5": "STAR", "6": "SCER", "7": "SCAN", "8": "SCAR", "9": "SCAC"},
#     "SID": {"1": "Sd(a"}, "VEL": {"1": "Vel"}, "XK": {"1": "kx"},
#     "YT": {"1": "Yt(a", "2": "Yt(b", "3": "YTEG", "4": "YTLI", "5": "YTOT"},
# }


class TestAntigenProfileComparison(unittest.TestCase):
    def _cmp(self, num: str, alpha: str, system: str, strict: bool = True) -> bool:
        return compare_antigen_profiles(num, alpha, ANTIGEN_MAP, system, strict=strict)

    def test_basic_positive_negative(self):
        self.assertTrue(self._cmp("RH:-2,3", "C-,E+", "RH"))
        self.assertFalse(self._cmp("RH:-2,3", "C-,E-", "RH"))

    def test_case_sensitivity_in_alpha_name(self):
        self.assertTrue(self._cmp("RH:-2,5", "C-,e+", "RH"))
        self.assertFalse(self._cmp("RH:-2,3", "C-,e+", "RH"))

    def test_alpha_name_validity(self):
        self.assertTrue(self._cmp("SC:-4,5", "Rd-,STAR+", "SC"))
        self.assertFalse(
            self._cmp("SC:-4,5", "rd-,STAR+", "SC")
        )  # 'rd' is not the canonical 'Rd'

    def test_correctly_signed_alpha_expression(self):
        self.assertTrue(self._cmp("RHAG:3,5", "DSLK+,Kg+", "RHAG"))
        self.assertTrue(self._cmp("OK:2,3", "OKGV+,OKVM+", "OK"))

    def test_order_sensitivity(self):
        self.assertTrue(self._cmp("OK:2,3", "OKGV+,OKVM+", "OK"))
        self.assertTrue(
            self._cmp("OK:2,3", "OKVM+,OKGV+", "OK")
        )  # Set comparison, order insensitive

    def test_modifier_weak(self):
        self.assertTrue(self._cmp("RH:4w", "c+weak", "RH"))
        self.assertFalse(self._cmp("RH:4w", "c+", "RH"))

    def test_modifier_partial(self):
        self.assertTrue(self._cmp("RH:5p", "e+partial", "RH"))

    def test_modifier_negative_transition(self):
        self.assertTrue(self._cmp("RH:5n", "e+positive_to_neg", "RH"))
        self.assertTrue(self._cmp("RH:5n", "e+negative", "RH"))
        self.assertFalse(self._cmp("RH:5n", "e+positive", "RH"))

    def test_modifier_monoclonal(self):
        self.assertTrue(self._cmp("RH:58m", "CELO+monoclonal", "RH"))

    def test_modifier_inferred(self):
        self.assertTrue(self._cmp("RH:31i", "hrB+inferred", "RH"))

    def test_modifier_robust(self):
        self.assertTrue(self._cmp("VEL:1r", "Vel+robust", system="VEL"))

    def test_modifier_strong(self):
        self.assertTrue(self._cmp("VEL:1s", "Vel+strong", system="VEL"))

    def test_modifier_very_weak(self):
        ANTIGEN_MAP["KEL"]["2"] = "k"
        self.assertTrue(self._cmp("KEL:2v", "k+very_weak", system="KEL"))
        self.assertTrue(self._cmp("KEL:2v", "k+v", system="KEL"))

    def test_modifier_weak_partial(self):
        self.assertTrue(self._cmp("RH:4wp", "c+weak_partial", "RH"))
        self.assertTrue(self._cmp("RH:4wp", "c+partial weak", "RH"))
        self.assertTrue(self._cmp("RH:4wp", "c+wp", "RH"))

    def test_modifier_weak_partial_negative(self):
        self.assertTrue(self._cmp("RH:5pwn", "e+partial_weak_to_neg", "RH"))
        self.assertTrue(self._cmp("RH:5pwn", "e+pwn", "RH"))

    def test_real_world_examples(self):
        self.assertTrue(
            self._cmp("RH:-2,-3,4wp,5wp", "C-,E-,c+weak_partial,e+weak_partial", "RH")
        )
        self.assertTrue(
            self._cmp(
                "RH:-2,-3,4,5n,-18,-19,49w",
                "C-,E-,c+,e+positive_to_neg,Hr-,hrS-,STEM+weak",
                "RH",
            )
        )
        self.assertFalse(
            self._cmp(
                "RH:-2,-3,4,5n,-18,-19,49",
                "C-,E-,c+,e+positive_to_neg,Hr-,hrS-,STEM+weak",
                "RH",
            )
        )
        self.assertFalse(
            self._cmp(
                "RH:-2,-3,4,5n,-18,-19,49w",
                "C-,E-,c+,e+positive,Hr-,hrS-,STEM+weak",
                "RH",
            )
        )

    def test_extra_alpha_antigen_strict(self):
        self.assertFalse(self._cmp("RH:-2,3", "C-,E+,hrB+", "RH", strict=True))

    def test_extra_alpha_antigen_lenient(self):
        self.assertTrue(self._cmp("RH:-2,3", "C-,E+,hrB+", "RH", strict=False))

    def test_missing_required_modifier(self):
        self.assertFalse(self._cmp("RH:5wp", "e+weak", "RH"))


ANTIGEN_MAP: dict[str, dict[str, str]] = {
    "ABCC1": {"1": "WLF"},
    "ATP11C": {"1": "LIL"},
    "AUG": {"1": "AUG1", "2": "At(a", "3": "ATML", "4": "ATAM"},
    "CD59": {"1": "CD59.1"},
    "CO": {"1": "Co(a", "2": "Co(b", "3": "Co3", "4": "Co4"},
    "CROM": {
        "1": "Cr(a",
        "10": "UMC",
        "11": "GUTI",
        "12": "SERF",
        "13": "ZENA",
        "14": "CROV",
        "15": "CRAM",
        "16": "CROZ",
        "17": "CRUE",
        "18": "CRAG",
        "19": "CROK",
        "2": "Tc(a",
        "20": "CORS",
        "3": "Tc(b",
        "4": "Tc(c",
        "5": "Dr(a",
        "6": "Es(a",
        "7": "IFC",
        "8": "WES(a",
        "9": "WES(b",
    },
    "CTL2": {"1": "VER", "2": "RIF", "3": "Cs(a", "4": "Cs(b", "5": "BROS"},
    "DI": {
        "1": "Di(a",
        "10": "Bp(a",
        "11": "Mo(a",
        "12": "Hg(a",
        "13": "Vg(a",
        "14": "Sw(a",
        "15": "BOW",
        "16": "NFLD",
        "17": "Jn(a",
        "18": "KREP",
        "19": "Tr(a",
        "2": "Di(b",
        "20": "Fr(a",
        "21": "SW1",
        "22": "DISK",
        "23": "DIST",
        "3": "Wr(a",
        "4": "Wr(b",
        "5": "Wd(a",
        "6": "Rb(a",
        "7": "WARR",
        "8": "ELO",
        "9": "Wu",
    },
    "DO": {
        "1": "Do(a",
        "10": "DODE",
        "2": "Do(b",
        "3": "Gy(a",
        "4": "Hy",
        "5": "Jo(a",
        "6": "DOYA",
        "7": "DOMR",
        "8": "DOLG",
        "9": "DOLC",
    },
    "EMM": {"1": "Emm"},
    "ER": {"1": "Er(a", "2": "Er(b", "3": "Er3", "4": "ERSA", "5": "ERAMA"},
    "FORS": {"1": "FORS"},
    "FY": {"1": "Fy(a", "2": "Fy(b", "3": "Fy3", "5": "Fy5", "6": "Fy6"},
    "GE": {
        "10": "GEPL",
        "11": "GEAT",
        "12": "GETI",
        "13": "GECT",
        "14": "GEAR",
        "2": "Ge2",
        "3": "Ge3",
        "4": "Ge4",
        "5": "Wb",
        "6": "Ls(a",
        "7": "An(a",
        "8": "Dh(a",
        "9": "GEIS",
    },
    "GIL": {"1": "GIL"},
    "GLOB": {"1": "P"},
    "I": {"1": "I"},
    "IN": {
        "1": "In(a",
        "2": "In(b",
        "3": "INFI",
        "4": "INJA",
        "5": "INRA",
        "6": "INSL",
    },
    "JK": {"1": "Jk(a", "2": "Jk(b", "3": "Jk3"},
    "JMH": {
        "1": "JMH",
        "2": "JMHK",
        "3": "JMHL",
        "4": "JMHG",
        "5": "JMHM",
        "6": "JMHQ",
        "7": "JMHN",
        "8": "JMHA",
    },
    "KANNO": {"1": "KANNO1"},
    "KEL": {
        "1": "K",
        "10": "Ul(a",
        "11": "K11",
        "12": "K12",
        "13": "K13",
        "14": "K14",
        "16": "K16",
        "17": "K17",
        "18": "K18",
        "19": "K19",
        "2": "k",
        "20": "Km",
        "21": "Kp(c",
        "22": "K22",
        "23": "K23",
        "24": "K24",
        "25": "VLAN",
        "26": "TOU",
        "27": "RAZ",
        "28": "VONG",
        "29": "KALT",
        "3": "Kp(a",
        "30": "KTIM",
        "31": "KYO",
        "32": "KUCI",
        "33": "KANT",
        "34": "KASH",
        "35": "KELP",
        "36": "KETI",
        "37": "KHUL",
        "38": "KYOR",
        "39": "KEAL",
        "4": "Kp(b",
        "40": "KHIZ",
        "41": "KHOZ",
        "5": "Ku",
        "6": "Js(a",
        "7": "Js(b",
    },
    "KN": {
        "1": "Kn(a",
        "10": "KDAS",
        "11": "DACY",
        "12": "YCAD",
        "13": "KNMB",
        "2": "Kn(b",
        "3": "McC(a",
        "4": "Sl1",
        "5": "Yk(a",
        "6": "McC(b",
        "7": "Vil",
        "8": "Sl3",
        "9": "KCAM",
    },
    "LU": {
        "1": "Lu(a",
        "11": "Lu11",
        "12": "Lu12",
        "13": "Lu13",
        "14": "Lu14",
        "16": "Lu16",
        "17": "Lu17",
        "18": "Au(a",
        "19": "Au(b",
        "2": "Lu(b",
        "20": "Lu20",
        "21": "Lu21",
        "22": "LURC",
        "23": "LUIT",
        "24": "LUGA",
        "25": "LUAC",
        "26": "LUBI",
        "27": "LUYA",
        "28": "LUNU",
        "29": "LURA",
        "3": "Lu3",
        "30": "LUOM",
        "4": "Lu4",
        "5": "Lu5",
        "6": "Lu6",
        "7": "Lu7",
        "8": "Lu8",
        "9": "Lu9",
    },
    "LW": {"5": "LWa", "6": "LWab", "7": "LWb", "8": "LWEM"},
    "MAM": {"1": "MAM"},
    "MNS": {
        "1": "M",
        "10": "Mur",
        "11": "Mg",
        "12": "Vr",
        "13": "Me",
        "14": "Mt(a",
        "15": "St(a",
        "16": "Ri(a",
        "17": "Cl(a",
        "18": "Ny(a",
        "19": "Hut",
        "2": "N",
        "20": "Hil",
        "21": "Mv",
        "22": "Far",
        "23": "sD",
        "24": "Mit",
        "25": "Dantu",
        "26": "Hop",
        "27": "Nob",
        "28": "En(a",
        "29": "ENKT",
        "3": "S",
        "30": "`N'",
        "31": "Or",
        "32": "DANE",
        "33": "TSEN",
        "34": "MINY",
        "35": "MUT",
        "36": "SAT",
        "37": "ERIK",
        "38": "Os(a",
        "39": "ENEP",
        "4": "s",
        "40": "ENEH",
        "41": "HAG",
        "42": "ENAV",
        "43": "MARS",
        "44": "ENDA",
        "45": "ENEV",
        "46": "MNTD",
        "47": "SARA",
        "48": "KIPP",
        "49": "JENU",
        "5": "U",
        "50": "SUMI",
        "6": "He",
        "7": "Mia",
        "8": "Mc",
        "9": "Vw",
    },
    "OK": {"1": "Ok(a", "2": "OKGV", "3": "OKVM"},
    "PEL": {"1": "PEL"},
    "RAPH": {"1": "MER2"},
    "RH": {
        "10": "V",
        "11": "Ew",
        "12": "G",
        "17": "Hro",
        "18": "Hr",
        "19": "hrS",
        "2": "C",
        "20": "VS",
        "21": "CG",
        "22": "CE",
        "23": "Dw",
        "26": "c_like",
        "27": "cE",
        "28": "hrH",
        "29": "Rh29",
        "3": "E",
        "30": "Goa",
        "31": "hrB",
        "32": "Rh32",
        "33": "Rh33",
        "34": "HrB",
        "35": "Rh35",
        "36": "Bea",
        "37": "Evans",
        "39": "Rh39",
        "4": "c",
        "40": "Tar",
        "41": "Rh41",
        "42": "Rh42",
        "43": "Crawford",
        "44": "Nou",
        "45": "Riv",
        "46": "Sec",
        "47": "Dav",
        "48": "JAL",
        "49": "STEM",
        "5": "e",
        "50": "FPTT",
        "51": "MAR",
        "52": "BARC",
        "53": "JAHK",
        "54": "DAK",
        "55": "LOCR",
        "56": "CENR",
        "57": "CEST",
        "58": "CELO",
        "59": "CEAG",
        "6": "f",
        "60": "PARG",
        "61": "CEVF",
        "62": "CEWA",
        "63": "CETW",
        "7": "Ce",
        "8": "Cw",
        "9": "Cx",
    },
    "RHAG": {
        "1": "Duclos",
        "2": "Ol(a",
        "3": "DSLK",
        "5": "Kg",
        "6": "SHER",
        "7": "THIN",
    },
    "RHD": {"1": "D"},
    "SC": {
        "1": "Sc1",
        "2": "Sc2",
        "3": "Sc3",
        "4": "Rd",
        "5": "STAR",
        "6": "SCER",
        "7": "SCAN",
        "8": "SCAR",
        "9": "SCAC",
    },
    "SID": {"1": "Sd(a"},
    "VEL": {"1": "Vel"},
    "XK": {"1": "kx"},
    "YT": {"1": "Yt(a", "2": "Yt(b", "3": "YTEG", "4": "YTLI", "5": "YTOT"},
}

if __name__ == "__main__":
    unittest.main()



import os
import re
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Optional, Set, Tuple

import pandas as pd
from loguru import logger
from reportlab.lib import colors
from reportlab.lib.pagesizes import letter
from reportlab.lib.styles import ParagraphStyle, getSampleStyleSheet
from reportlab.lib.units import inch
from reportlab.pdfgen import canvas
from reportlab.platypus import (
    BaseDocTemplate,
    Flowable,
    Frame,
    PageTemplate,
    Paragraph,
    Spacer,
    Table,
    TableStyle,
)

from rbceq2.core_logic.constants import DB_VERSION, VERSION

# --- Helper Functions ---


def _normalize_sample_id(sample_id: str) -> str:
    """Removes common suffixes from sample IDs for consistent matching.

    Args:
        sample_id: The original sample identifier string.

    Returns:
        The normalized sample identifier string.
    """
    # Ensure input is string before regex
    sample_id_str = str(sample_id)
    # Remove the common long suffix first, including optional .vcf
    normalized = re.sub(
        r"_GRCh38_1_22_v4\.2\.1_benchmark_filtered(\.vcf)?$", "", sample_id_str
    )
    # Example: remove just .vcf if the long suffix wasn't present
    normalized = re.sub(r"\.vcf$", "", normalized)
    # Add any other normalization rules here if needed
    return normalized


def _format_cell_content(text: Optional[Any], separator: Optional[str] = None) -> str:
    """Formats cell text for PDF tables, handling N/A, line breaks, and HTML escaping.

    Args:
        text: The content of the cell (can be string, number, NaN, None).
        separator: The character(s) used to split the text for line breaks (e.g., ',').
                   If None, no splitting occurs, only escaping and N/A handling.

    Returns:
        A string formatted for use in a ReportLab Paragraph, with <br/> for line breaks
        and escaped HTML characters, or "N/A" if the input is missing/empty.
    """
    if pd.isna(text) or text == "" or text is None:
        return "N/A"

    text_str = str(text)
    # Escape HTML special characters first to avoid breaking tags later
    # Using standard library for robust escaping
    import html

    escaped_text = html.escape(text_str)

    if separator:
        # Split the original unescaped text by the separator
        items = text_str.split(separator)
        # Escape *each item* after splitting and strip whitespace
        escaped_items = [html.escape(item.strip()) for item in items]
        # Join the individually escaped items with <br/>
        return "<br/>".join(escaped_items)
    else:
        # Return the fully escaped text if no separator provided
        return escaped_text


# --- Data Preparation ---


def _prepare_dataframes(
    df_genotypes: Optional[pd.DataFrame],
    df_pheno_alpha: Optional[pd.DataFrame],
    df_pheno_num: Optional[pd.DataFrame],
) -> Tuple[Dict[str, Optional[pd.DataFrame]], Set[str], Dict[str, str]]:
    """Normalizes sample IDs in DataFrames and identifies unique samples.

    Adds a 'SampleID_Normalized' column to each input DataFrame.

    Args:
        df_genotypes: DataFrame with genotype data. Index should be SampleID.
        df_pheno_alpha: DataFrame with alphanumeric phenotype data. Index should be SampleID.
        df_pheno_num: DataFrame with numeric phenotype data. Index should be SampleID.

    Returns:
        A tuple containing:
        - A dictionary mapping data type ('genotype', 'alpha', 'numeric') to the
          corresponding DataFrame (or None if input was None/empty). DataFrames
          will have the 'SampleID_Normalized' column added and index reset.
        - A set of unique normalized sample IDs found across all DataFrames.
        - A dictionary mapping normalized sample IDs to the first encountered
          original sample ID.
    """
    dfs = {"genotype": df_genotypes, "alpha": df_pheno_alpha, "numeric": df_pheno_num}
    processed_dfs: Dict[str, Optional[pd.DataFrame]] = {}
    original_id_map: Dict[str, str] = {}  # Map normalized ID -> first original ID
    all_normalized_ids: Set[str] = set()

    for df_type, df in dfs.items():
        if df is not None and not df.empty:
            df_processed = df.copy()
            original_ids = df_processed.index.astype(
                str
            ).tolist()  # Get original IDs before potentially changing index
            df_processed["SampleID_Normalized"] = df_processed.index.map(
                _normalize_sample_id
            )

            all_normalized_ids.update(df_processed["SampleID_Normalized"].unique())

            # Store mapping: normalized -> original
            for i, norm_id in enumerate(df_processed["SampleID_Normalized"]):
                if norm_id not in original_id_map:
                    original_id_map[norm_id] = original_ids[
                        i
                    ]  # Map normalized to the first original ID encountered

            processed_dfs[df_type] = df_processed
        else:
            processed_dfs[df_type] = None

    return processed_dfs, all_normalized_ids, original_id_map


# --- PDF Styling and Content Generation ---


# def _setup_styles() -> Dict[str, ParagraphStyle]:
#     """Creates and returns a dictionary of ReportLab ParagraphStyle objects.

#     Returns:
#         A dictionary mapping style names (e.g., 'body', 'title', 'warning')
#         to their corresponding ParagraphStyle objects.
#     """
#     styles = getSampleStyleSheet()
#     # Reduce body text size slightly for potentially better table fit
#     body_style = ParagraphStyle(
#         name="BodyTextSmall", parent=styles["BodyText"], fontSize=9
#     )
#     table_cell_style = ParagraphStyle(
#         name="TableCell",
#         parent=body_style,
#         leading=11,
#         fontSize=8,  # Smaller font, tighter leading for cells
#     )
#     footer_style = ParagraphStyle(
#         name="FooterStyle",
#         parent=styles["Normal"],
#         alignment=1,
#         fontSize=7,
#         textColor=colors.grey,  # Centered footer
#     )

#     custom_styles = {
#         "body": body_style,
#         "heading": styles["h2"],
#         "title": styles["h1"],
#         "report_title": ParagraphStyle(
#             name="ReportTitle",
#             parent=styles["h1"],
#             alignment=1,
#             fontSize=16,  # Centered H1
#         ),
#         "warning": ParagraphStyle(
#             name="Warning",
#             parent=styles["Heading1"],
#             textColor=colors.red,
#             alignment=1,
#             fontSize=14,
#         ),
#         "table_header": ParagraphStyle(
#             name="TableHeader",
#             parent=styles["BodyText"],
#             fontName="Helvetica-Bold",
#             fontSize=8,  # Smaller header
#         ),
#         "table_cell": table_cell_style,
#         "footer": footer_style,
#     }
#     return custom_styles

def _setup_styles() -> Dict[str, ParagraphStyle]:
    """Creates and returns a dictionary of ReportLab ParagraphStyle objects.

    Returns:
        A dictionary mapping style names (e.g., 'body', 'title', 'warning')
        to their corresponding ParagraphStyle objects.
    """
    styles = getSampleStyleSheet()
    body_style = ParagraphStyle(name='BodyTextSmall', parent=styles['BodyText'], fontSize=9)

    table_cell_style = ParagraphStyle(
        name="TableCell",
        parent=body_style,
        fontSize=7,      
        leading=9,      
    )
    # ****************************************************************

    footer_style = ParagraphStyle(
        name='FooterStyle', parent=styles['Normal'], alignment=1, fontSize=7, textColor=colors.grey # Centered footer
    )

    custom_styles = {
        "body": body_style,
        "heading": styles["h2"],
        "title": styles["h1"],
        "report_title": ParagraphStyle(
            name="ReportTitle", parent=styles["h1"], alignment=1, fontSize=16 # Centered H1
        ),
        "warning": ParagraphStyle(
            name="Warning", parent=styles["Heading1"], textColor=colors.red, alignment=1, fontSize=14
        ),
        "table_header": ParagraphStyle(
            # Keep header size reasonable or reduce slightly if needed too
            name="TableHeader", parent=styles["BodyText"], fontName="Helvetica-Bold", fontSize=8
        ),
        "table_cell": table_cell_style, # Use the modified style
        "footer": footer_style,
    }
    return custom_styles

def _add_page_footer(
    canvas: canvas.Canvas,
    doc: BaseDocTemplate,
    styles: Dict[str, ParagraphStyle],
    UUID: str,
) -> None:
    """Draws the footer on each page."""
    canvas.saveState()
    timestamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    footer_text = (
        f"Generated: {timestamp} | Code: v{VERSION} | DB: {DB_VERSION} | UUID: {UUID}"
    )
    p = Paragraph(footer_text, styles["footer"])
    w, h = p.wrap(doc.width, doc.bottomMargin)  # Use doc.width available space
    p.drawOn(
        canvas, doc.leftMargin, doc.bottomMargin - h - 5
    )  # Position below bottom margin
    canvas.restoreState()


def _create_report_header(
    story: List[Flowable], original_id_display: str, styles: Dict[str, ParagraphStyle]
) -> None:
    """Adds standard header elements to the PDF story.

    Args:
        story: The list of ReportLab Flowables to append to.
        original_id_display: The sample ID to display in the header.
        styles: Dictionary containing the required ParagraphStyles.
    """
    story.append(Paragraph("RBCeq2", styles["report_title"]))
    story.append(Spacer(1, 0.1 * inch))
    story.append(Paragraph("Not for clinical use", styles["warning"]))
    story.append(Spacer(1, 0.2 * inch))
    story.append(Paragraph(f"Sample ID: {original_id_display}", styles["heading"]))
    story.append(Spacer(1, 0.1 * inch))  # Reduced space before table title


def _get_data_for_sample(
    norm_id: str, processed_dfs: Dict[str, Optional[pd.DataFrame]]
) -> Tuple[Dict[str, Optional[pd.Series]], Set[str]]:
    """Retrieves data rows for a specific normalized sample ID from processed dfs.

    Args:
        norm_id: The normalized sample ID to retrieve data for.
        processed_dfs: Dictionary mapping data type to processed DataFrame
                       (must contain 'SampleID_Normalized' column).

    Returns:
        A tuple containing:
        - A dictionary mapping data type ('genotype', 'alpha', 'numeric')
          to the pd.Series containing the sample's data (or None if not found).
          The Series index will be the gene/system names.
        - A set of all unique gene/system keys found for this sample across
          all data types.
    """
    sample_data: Dict[str, Optional[pd.Series]] = {}
    all_keys_for_sample: Set[str] = set()

    for df_type, df in processed_dfs.items():
        row_data: Optional[pd.Series] = None
        if df is not None:
            # Find rows matching the normalized ID
            matching_rows = df[df["SampleID_Normalized"] == norm_id]
            if not matching_rows.empty:
                # Select the first match (assuming one row per normalized ID per df type)
                # Drop the helper column to get only gene/system data
                row_data = matching_rows.iloc[0].drop(
                    "SampleID_Normalized", errors="ignore"
                )  # Ignore error if column somehow missing
                all_keys_for_sample.update(row_data.index)

        sample_data[df_type] = row_data

    return sample_data, all_keys_for_sample


def _create_consolidated_table(
    sample_data: Dict[str, Optional[pd.Series]],
    all_keys_for_sample: Set[str],
    styles: Dict[str, ParagraphStyle],
) -> Flowable:
    """Creates the main data table combining genotype and phenotypes.

    Args:
        sample_data: Dict mapping data type to the sample's data Series.
        all_keys_for_sample: Set of all gene/system keys for this sample.
        styles: Dictionary containing the required ParagraphStyles.

    Returns:
        A ReportLab Table object or a Paragraph indicating no data.
    """
    if not all_keys_for_sample:
        return Paragraph("No data available for this sample.", styles["body"])

    # Prepare table header row
    header_content = ["System/Gene", "Genotype", "Phenotype (Alpha)", "Phenotype (Num)"]
    # Wrap header text in Paragraphs using the header style
    header_row: List[Paragraph] = [
        Paragraph(text, styles["table_header"]) for text in header_content
    ]
    combined_table_data: List[List[Flowable]] = [header_row]

    # Sort keys alphabetically for consistent row order
    sorted_keys = sorted(list(all_keys_for_sample))

    for key in sorted_keys:
        # Safely get data using .get(key, default_value) for robustness
        geno_val = (
            sample_data["genotype"].get(key, None)
            if sample_data["genotype"] is not None
            else None
        )
        alpha_val = (
            sample_data["alpha"].get(key, None)
            if sample_data["alpha"] is not None
            else None
        )
        num_val = (
            sample_data["numeric"].get(key, None)
            if sample_data["numeric"] is not None
            else None
        )

        # Format values for the table cells
        # Check data examples to decide on separators if needed
        # Assuming genotype might contain '\n' which _format_cell_content should handle
        formatted_geno = _format_cell_content(geno_val, separator="\n")
        # Use newline as separator if present in data
        formatted_alpha = _format_cell_content(alpha_val)
        # No separator assumed for alpha
        formatted_num = _format_cell_content(
            num_val
        )  # No separator assumed for numeric

        # Create Paragraphs for body cells using the cell style
        combined_table_data.append(
            [
                Paragraph(str(key), styles["table_cell"]),
                Paragraph(formatted_geno, styles["table_cell"]),
                Paragraph(formatted_alpha, styles["table_cell"]),
                Paragraph(formatted_num, styles["table_cell"]),
            ]
        )

    # Slightly adjusted widths
    col_widths = [1.5 * inch, 2.5 * inch, 2.0 * inch, 1.5 * inch]

    # Create the table
    combined_table = Table(combined_table_data, colWidths=col_widths, splitByRow=1)

    # Apply styling (using your original style block)
    combined_table.setStyle(
        TableStyle(
            [
                ("BACKGROUND", (0, 0), (-1, 0), colors.grey),
                ("TEXTCOLOR", (0, 0), (-1, 0), colors.whitesmoke),
                ("ALIGN", (0, 0), (-1, 0), "CENTER"),  # Header alignment
                ("ALIGN", (0, 1), (0, -1), "LEFT"),  # Gene name left aligned
                ("ALIGN", (1, 1), (-1, -1), "LEFT"),  # Other columns LEFT aligned
                ("VALIGN", (0, 0), (-1, -1), "MIDDLE"),  # Vertical alignment
                ("FONTNAME", (0, 0), (-1, 0), "Helvetica-Bold"),
                ("BOTTOMPADDING", (0, 0), (-1, 0), 8),  # Padding for header
                ("TOPPADDING", (0, 0), (-1, 0), 4),
                ("BOTTOMPADDING", (0, 1), (-1, -1), 4),  # Padding for cells
                ("TOPPADDING", (0, 1), (-1, -1), 4),
                ("BACKGROUND", (0, 1), (-1, -1), colors.beige),
                ("GRID", (0, 0), (-1, -1), 1, colors.black),
                ("LEFTPADDING", (0, 0), (-1, -1), 4),
                ("RIGHTPADDING", (0, 0), (-1, -1), 4),
            ]
        )
    )
    return combined_table


def _generate_pdf_report_for_sample(
    norm_id: str,
    original_id_display: str,
    processed_dfs: Dict[str, Optional[pd.DataFrame]],
    styles: Dict[str, ParagraphStyle],
    output_dir: Path,
    UUID: str,
) -> None:
    """Generates and saves a single PDF report for one sample.

    Args:
        norm_id: The normalized sample ID.
        original_id_display: The original sample ID to use for display and filename.
        processed_dfs: Dict mapping data type to the processed DataFrame.
        styles: Dictionary of required ParagraphStyles.
        output_dir: The directory to save the PDF file.
    """
    # Sanitize original ID for filename
    safe_original_id = "".join(
        c if c.isalnum() or c in ("_", "-") else "_" for c in original_id_display
    )
    pdf_filename = os.path.join(output_dir, f"{safe_original_id}_BloodGroupReport.pdf")

    # Use BaseDocTemplate to allow page footers
    doc = BaseDocTemplate(
        pdf_filename,
        pagesize=letter,  # Use letter size as per your original code
        leftMargin=0.75 * inch,
        rightMargin=0.75 * inch,
        topMargin=0.75 * inch,  # Increased top margin slightly
        bottomMargin=0.75 * inch,
    )

    # Define the frame for content flow
    frame = Frame(doc.leftMargin, doc.bottomMargin, doc.width, doc.height, id="normal")
    # Create a PageTemplate that uses the footer function
    template = PageTemplate(
        id="main",
        frames=[frame],
        onPage=lambda canvas, doc: _add_page_footer(canvas, doc, styles, UUID),
    )
    doc.addPageTemplates([template])

    story: List[Flowable] = []

    # 1. Add Header (after PageTemplate setup)
    _create_report_header(story, original_id_display, styles)

    # 2. Get Data for this sample
    sample_data, all_keys_for_sample = _get_data_for_sample(norm_id, processed_dfs)

    # 3. Add Main Table Title
    story.append(
        Paragraph("Blood Group Genotype & Predicted Phenotype", styles["heading"])
    )
    story.append(Spacer(1, 0.1 * inch))

    # 4. Create and add the data table (or "No data" message)
    table_flowable = _create_consolidated_table(
        sample_data, all_keys_for_sample, styles
    )
    story.append(table_flowable)
    # Removed spacer after table to maximize space

    # 5. Build the PDF
    try:
        doc.build(story)
    except Exception as e:
        logger.error(f"ERROR generating PDF for {norm_id} ({original_id_display}): {e}")


# --- Main Orchestration Function ---


def generate_all_reports(
    df_genotypes: Optional[pd.DataFrame],
    df_pheno_alpha: Optional[pd.DataFrame],
    df_pheno_num: Optional[pd.DataFrame],
    output_name: Path,
    UUID: str,
) -> None:
    """
    Generates consolidated PDF blood group reports for all samples found in input DataFrames.

    Args:
        df_genotypes: DataFrame with genotype data. Index must be SampleID.
        df_pheno_alpha: DataFrame with alphanumeric phenotype data. Index must be SampleID.
        df_pheno_num: DataFrame with numeric phenotype data. Index must be SampleID.
        output_dir: Directory where the generated PDF files will be saved.
    """

    # 1. Prepare Data & Get Unique IDs
    processed_dfs, all_normalized_ids, original_id_map = _prepare_dataframes(
        df_genotypes, df_pheno_alpha, df_pheno_num
    )

    if not all_normalized_ids:
        logger.warning("No samples found in the input DataFrames. Exiting.")
        return

    num_samples = len(all_normalized_ids)
    logger.info(f"Found {num_samples} unique normalized sample IDs.")
    output_dir = f"{output_name}_PDFs"
    # 2. Create output directory
    os.makedirs(output_dir, exist_ok=True)
    logger.info(f"Reports will be saved to '{os.path.abspath(output_dir)}'.")

    # 3. Setup Styles
    styles = _setup_styles()

    # 4. Iterate Through Each Sample and Generate PDF
    for norm_id in sorted(list(all_normalized_ids)):
        original_id_display = original_id_map.get(
            norm_id, norm_id
        )  # Fallback to norm_id if mapping failed

        try:
            _generate_pdf_report_for_sample(
                norm_id, original_id_display, processed_dfs, styles, output_dir, UUID
            )
        except Exception as e:
            # Catch potential errors during individual PDF generation
            logger.error(
                f"Unhandled exception during PDF generation for {original_id_display}: {e}"
            )



import itertools
import operator
from collections import defaultdict
from dataclasses import dataclass
from functools import partial
from typing import Any, Protocol
from loguru import logger

from rbceq2.core_logic.alleles import Allele, BloodGroup, Pair
from rbceq2.core_logic.constants import AlleleState
from rbceq2.core_logic.utils import (
    Zygosity,
    apply_to_dict_values,
    check_available_variants,
    chunk_geno_list_by_rank,
    get_non_refs,
)
from rbceq2.db.db import Db
from rbceq2.IO.vcf import VCF
from icecream import ic



def raw_results(db: Db, vcf: VCF, exclude: list[str]) -> dict[str, list[Allele]]:
    """Generate raw results from database alleles and VCF data based on phasing
    information.

    Args:
        db (Db): The database containing allele definitions and methods to generate
        them.
        vcf (VCF): The VCF data containing variants and possibly phasing information.

    Returns:
        Dict[str, List[Allele]]: A dictionary mapping blood groups to lists of Allele
        objects.
    """
    res: dict[str, list[Allele]] = defaultdict(list)

    for allele in db.make_alleles():
        if any(x in allele.genotype for x in exclude):
            continue
        if all(var in vcf.variants for var in allele.defining_variants):
            res[allele.blood_group].append(allele)

    return res


@apply_to_dict_values
def add_phasing(
    bg: BloodGroup,
    phased: bool,
    variant_metrics: dict[str, dict[str, str]],
    phase_sets: dict[str, dict[int, tuple[int, int]]],
) -> BloodGroup:
    """Add phasing information to alleles in a BloodGroup.

    The Biological Scenario (gemini 2.5 pro)
    Imagine a gene called GENEX. In an individual, we find two different
    heterozygous variants within this gene:
    At position chr1:1,000,000, there's a A > G variant.
    At position chr1:1,002,500, there's a C > T variant.
    Since this person is diploid, they have two copies of GENEX (one
    maternal, one paternal). The critical question that phasing answers
    is: Are the two alternate alleles (G and T) on the same copy of the
    chromosome, or on different copies?
    There are two possibilities:
    Scenario A (Cis configuration): The alternate alleles are on the same chromosome.
    Maternal Chromosome: ---G---T---
    Paternal Chromosome: ---A---C--- (reference alleles)
    Scenario B (Trans configuration): The alternate alleles are on opposite chromosomes.
    Maternal Chromosome: ---G---C---
    Paternal Chromosome: ---A---T---

    How This Looks in the VCF
    A phasing algorithm (using long reads, parental data, or statistical inference)
    will try to determine whether we are in Scenario A or B. If it succeeds,
    it will report both variants in the same phase set.
    Let's assign the PS (Phase Set) identifier 12345 to this phased block.
    VCF for Scenario A (Cis):
    The two alternate alleles (1) are on the same haplotype.
    Generated vcf
    #CHROM  POS        ID  REF ALT ... FORMAT     SAMPLE1
    chr1    1000000    .   A   G   ... GT:PS      0|1:12345
    chr1    1002500    .   C   T   ... GT:PS      0|1:12345
    Vcf
    PS:12345: Both variants are in the same phase set.
    0|1 and 0|1: This tells us that the allele designated '1' (the ALT allele)
    for the first variant is on the same chromosome as the allele designated
    '1' for the second variant. Haplotype 1 is G-T, Haplotype 0 is A-C.
    VCF for Scenario B (Trans):
    The two alternate alleles (1) are on opposite haplotypes.
    Generated vcf
    #CHROM  POS        ID  REF ALT ... FORMAT     SAMPLE1
    chr1    1000000    .   A   G   ... GT:PS      0|1:12345
    chr1    1002500    .   C   T   ... GT:PS      1|0:12345
    Vcf

    PS:12345: They are still in the same phase set. This is the key point. The
    PS just tells you "the relationship between these variants is known."

    0|1 and 1|0: The GT fields now describe the trans relationship. For the first
    variant, the alternate allele (G) is on Haplotype 1. For the second variant,
    the alternate allele (T) is on Haplotype 0. Haplotype 1 is G-C, Haplotype
    0 is A-T.

    When Would They Be in Different Phase Sets?
    Variants within the same gene would only appear in different phase sets if
    the phasing process failed to connect them.


    If 'phased' is True, updates the alleles in the given 'bg' object by assigning
    phase sets (PS) from 'variant_metrics'. Alleles containing reference-only variants
    are ignored or reduced in count.

    Args:
        bg (BloodGroup):
            The BloodGroup object whose alleles are to be phased.
        phased (bool):
            A flag indicating whether phasing is enabled or not.
        variant_metrics (dict[str, dict[str, str]]):
            A nested dictionary of metrics for each variant. The inner dictionary
            should include the phase set (PS) or other variant-related metrics.

    Returns:
        BloodGroup:
            The updated BloodGroup object with phased alleles, if 'phased' is True.
            Otherwise, the original BloodGroup object.

    Raises:
        AssertionError:
            If the number of updated alleles does not match the number of original alleles,
            or if the computed phases do not match the expected length based on 'defining_variants'.
    """

    def assign_ref_phase(
        current_variant: str,
        # bg_variant_pool: dict[str, str],
        # current_phase_pool: dict[str, str],
    ) -> str:
        """
        Assigns the correct phase to a reference variant at a heterozygous site.

        This function finds the corresponding alternate allele at the same genomic
        position, retrieves its phase, and returns the inverse phase for the
        reference allele.

        Args:
            variant: The identifier of the reference variant (e.g., '4:144120554_ref').
            bg_variant_pool: A dictionary mapping variant IDs to their zygosity.
            phase_pool: A dictionary mapping variant IDs to their phase string (e.g., '1|0').

        Returns:
            The calculated phase string for the reference variant (e.g., '0|1').

        Raises:
            ValueError: If the input variant is not a valid heterozygous reference variant,
                        if a corresponding alternate variant cannot be found, or if the
                        phase format is invalid.
        """
        zygosity = bg.variant_pool.get(current_variant)
        if zygosity == Zygosity.HOM:
            return "1/1"
        if "/" in phase_pool[current_variant]:
            return "unknown"
        # 1. Validate that the input variant is a heterozygous reference variant
        if not current_variant.endswith("_ref"):
            raise ValueError(
                f"Input variant '{current_variant}' must be a reference variant (ending in '_ref')."
            )

        if zygosity != Zygosity.HET:
            raise ValueError(
                f"Variant '{current_variant}' must be present and 'Heterozygous' in bg_variant_pool. "
                f"Found: {zygosity}"
            )

        # 2. Find the corresponding alternate allele variant at the same position
        position = current_variant.split("_")[0]
        partner_variant = None
        for key in phase_pool:
            # Match keys that start with the same position but are not the ref variant itself
            if key.startswith(position + "_") and key != current_variant:
                partner_variant = key
                break  # Found the partner, no need to continue searching
        if partner_variant is None:
            return phase_pool[current_variant]

        # 3. Get the phase of the partner variant
        partner_phase = phase_pool.get(partner_variant)
        if partner_phase is None:
            raise ValueError(
                f"Partner variant '{partner_variant}' found but has no entry in phase_pool."
            )

        # 4. Calculate the ref phase by "flipping" the partner's phase
        phase_parts = partner_phase.split("|")
        if len(phase_parts) != 2:
            raise ValueError(
                f"Invalid phase format for partner '{partner_variant}': '{partner_phase}'"
            )

        # The ref allele is on the opposite haplotype from the alt allele
        # e.g., if partner is 1|0, ref is 0|1
        ref_phase = f"{phase_parts[1]}|{phase_parts[0]}"

        return ref_phase

    def assign_ref_phase_set(current_variant):
        """ """

        def get_phase_set_for_loci() -> str:
            """Queries for a phase set ID given a locus.

            Args:
                loci (str): A locus in "CHROM:POS" format (e.g., "1:12345"
                            or "chr1:12345").

            Returns:
                int | None: The phase set ID if the locus falls within a known
                            phased block, otherwise None.
            """
            chrom, pos_str = current_variant.split("_")[0].split(":")
            pos = int(pos_str)

            # Normalize chromosome name to match internal representation (e.g. '1' not 'chr1')
            chrom = chrom.replace("chr", "")

            # Get all phase sets for the given chromosome
            chrom_phase_sets = phase_sets.get(chrom)
            if not chrom_phase_sets:
                return "unknown"

            # Check if the position falls within any of the phase set ranges
            for ps_id, (min_pos, max_pos) in chrom_phase_sets.items():
                if min_pos <= pos <= max_pos:
                    return str(ps_id)

            # If no matching phase set is found
            return "unknown"

        zygosity = bg.variant_pool.get(current_variant)
        if zygosity == Zygosity.HOM:
            return "."
        # 2. Find the corresponding alternate allele variant at the same position
        position = current_variant.split("_")[0]
        partner_variant = None
        for key in phase_set_pool:
            # Match keys that start with the same position but are not the ref variant itself
            if key.startswith(position + "_") and key != current_variant:
                partner_variant = key
                break  # Found the partner, no need to continue searching
        if partner_variant is None:
            return get_phase_set_for_loci()
        partner = phase_set_pool.get(partner_variant)
        if partner is not None:
            return partner
        else:
            return get_phase_set_for_loci()

    if phased:  # TODO enum for GT, PS etc
        phase_pool = {
            variant: variant_metrics[variant]["GT"] for variant in bg.variant_pool
        }
        phase_set_pool = {
            variant: variant_metrics[variant].get("PS") for variant in bg.variant_pool
        }
        phase_pool_ref_fixed = {}
        
        for variant, phase in phase_pool.items():
            if variant.endswith("_ref"):
                new_phase = assign_ref_phase(variant)
                phase_pool_ref_fixed[variant] = new_phase
            else:
                phase_pool_ref_fixed[variant] = phase
        phase_set_pool_ref_fixed = {}

        for variant, phase in phase_set_pool.items():
            if variant.endswith("_ref") or phase is None:
                new_phase = assign_ref_phase_set(variant)
                phase_set_pool_ref_fixed[variant] = new_phase
            else:
                phase_set_pool_ref_fixed[variant] = phase
        bg.variant_pool_phase = phase_pool_ref_fixed
        bg.variant_pool_phase_set = phase_set_pool_ref_fixed

    return bg


@apply_to_dict_values
def ABO_phasing(
    bg: BloodGroup,
    phased: bool,
) -> BloodGroup:
    # aboO_phases2 = set([]) I'm opting out of this path but if it ever gets revisited
    # do it at DB level #TODO look for vars that have only ever been observed in cis with
    # ABO*O, or never been observed with it and take phasing info from there. Best 
    # to actually phase indels though 
    """9:133257521(GRCh38) and 9:136132908 (GRCh37) _T_TC or _ref are pivotal for ABO
    calls but aren't always assigned a phase group by phasing algos

    some alleles are same/similar except for this locus

    ie HG04054 for 1kg data

    Allele
    genotype: ABO*B.01
    defining_variants:
                9:133255935_G_T 0|1
                9:133257521_T_TC 0/1
                9:133257486_T_C 1/1
                9:133255928_C_G 0|1
                9:133256074_G_A 0|1
                9:133256028_C_T 0|1
                9:133256205_G_C 0|1
                9:133255801_C_T 0|1
    weight_geno: 1000
    phenotype: . or B
    reference: False

    Allele
    genotype: ABO*O.01.41
    defining_variants:
                9:133257486_T_C 1/1
                9:133255928_C_G 0|1
                9:133256074_G_A 0|1
                9:133256028_C_T 0|1
                9:133256205_G_C 0|1
                9:133257521_ref unknown
                9:133255801_C_T 0|1

    this function will infer the phase of this locus from the phase of ABO*O.01
    specific variants
    """
    if not phased:
        return bg

    if bg.type != "ABO":
        return bg
    # 261delG
    c261delGs = [
            "9:133257521_ref",
            "9:133257521_T_TC",
            "9:136132908_ref",
            "9:136132908_T_TC",
        ]
    if any(
        bg.variant_pool.get(c261delG) == Zygosity.HOM
        for c261delG in c261delGs
    ):
        return bg

    aboO = set([])
    other = set([])
    for allele in bg.alleles[AlleleState.FILT]:
        if allele.genotype.startswith("ABO*O.01"):
            for variant in allele.defining_variants:
                aboO.add(variant)
        else:
            for variant in allele.defining_variants:
                other.add(variant)
    aboO_phases = set([])
    for variant in aboO.difference(other):
        if not variant.startswith(("9:133257521", "9:136132908")):
            phase = bg.variant_pool_phase[variant]
            aboO_phases.add(phase)
    # aboO_phases2 = set([]) I'm opting out of this path but if it ever gets revisited
    # do it at DB level #TODO look for vars that have only ever been observed in cis with
    # ABO*O, or never been observed with it and take phasing info from there. Best 
    # to actually phase indels though 
    # for variant in other.difference(aboO):
    #     if not variant.startswith(("9:133257521", "9:136132908")):
    #         phase = bg.variant_pool_phase[variant]
    #         aboO_phases2.add(phase)
    #ic(aboO_phases,aboO_phases2,aboO, other, aboO.difference(other), other.difference(aboO))    
    if len(aboO_phases) == 0:
        return bg #can't rescue ABO 
    if len(aboO_phases) > 1:
        #ic(aboO_phases,aboO, other, aboO.difference(other), other.difference(aboO))    
        return bg #can't rescue ABO 
    abo_phase = aboO_phases.pop()
    not_abo_phase = "1|0" if abo_phase == "0|1" else "0|1"
    new_phases = {}
    for variant, phase in bg.variant_pool_phase.items():
        if variant in c261delGs:
            if variant.endswith('_ref'):
                new_phases[variant] = abo_phase
            else:
                new_phases[variant] = not_abo_phase
        else:
            new_phases[variant] = phase
    bg.variant_pool_phase = new_phases

    return bg


@apply_to_dict_values
def make_variant_pool(bg: BloodGroup, vcf: VCF) -> BloodGroup:
    """Construct or update a variant pool for a BloodGroup from VCF data.

    This function traverses the alleles in the BloodGroup object, extracts reference
    information for each defining variant from the VCF, and combines these into a
    single dictionary (the variant pool).

    Args:
        bg (BloodGroup):
            The BloodGroup object to be updated with the new variant pool.
        vcf (VCF):
            The VCF object providing variant data.

    Returns:
        BloodGroup:
            The updated BloodGroup object, including the combined 'variant_pool' with
            reference data for each defining variant.

    Raises:
        KeyError:
            If a variant in 'bg.alleles[AlleleState.FILT]' is not found in 'vcf.variants'.
    """
    variant_pool = {}

    for allele in bg.alleles[AlleleState.FILT]:
        zygosity = {var: get_ref(vcf.variants[var]) for var in allele.defining_variants}
        variant_pool = variant_pool | zygosity
    bg.variant_pool = variant_pool

    return bg


def get_ref(ref_dict: dict[str, str]) -> str:
    """Determine the zygosity from a reference dictionary containing genotype
    information.

    Args:
        ref_dict (Dict[str, str]): A dictionary containing the genotype ("GT") and
        possibly other information.

    Returns:
        str: A string indicating the zygosity as 'Homozygous' or 'Heterozygous'.

    Raises:
        ValueError: If the genotype string does not conform to the expected format.

    The genotype string is expected to be in the format '0/1', '0|1', etc., where
    the delimiter can be '/' or '|'.
    A genotype of '0/0' or '1/1', etc., where both alleles are the same, will return
    'Homozygous'.
    A genotype of '0/1', '1/0', etc., will return 'Heterozygous'.
    """
    # 0/1:41,47:88:99:1080,0,1068:0.534:99
    ref_str = ref_dict["GT"]
    assert len(ref_str) == 3

    ref_str = ref_str.replace(".", "0")
    if ref_str[0] == ref_str[2]:
        return Zygosity.HOM
    return Zygosity.HET


@apply_to_dict_values
def get_genotypes(bg: BloodGroup) -> BloodGroup:
    """Generate genotype combinations for a given blood group from allele pairs.

    Args:
        bg (BloodGroup): The blood group object containing alleles.

    Returns:
        BloodGroup: The blood group object with updated genotypes based on allele
        combinations.

    This function processes 'pairs' and 'co_existing' alleles to create sorted genotype
    strings.
    """

    def make_list_of_lists(alleles):
        return [pair.genotypes for pair in alleles]

    if bg.alleles[AlleleState.CO] is not None:
        bg.genotypes = [
            "/".join(sorted(co))
            for co in make_list_of_lists(bg.alleles[AlleleState.CO])
        ]
    else:
        bg.genotypes = [
            "/".join(sorted(normal_pair))
            for normal_pair in make_list_of_lists(bg.alleles[AlleleState.NORMAL])
        ]

    return bg


def make_blood_groups(
    res: dict[str, list[Allele]], sample: str
) -> dict[str, BloodGroup]:
    """Create a dictionary of BloodGroup objects from allele data.

    Iterates through the 'res' mapping of blood group identifiers to lists of Allele
    objects, and constructs a new dictionary where each key is a blood group name and
    each value is a BloodGroup instance.

    Args:
        res (dict[str, list[Allele]]):
            A dictionary mapping blood group names to a list of Allele objects.
        sample (str):
            The sample identifier to be associated with each BloodGroup.

    Returns:
        dict[str, BloodGroup]:
            A dictionary mapping blood group identifiers to BloodGroup instances.
    """
    new_dict: dict[str, BloodGroup] = {}
    for blood_group, alleles in res.items():
        new_dict[blood_group] = BloodGroup(
            type=blood_group, alleles={AlleleState.RAW: alleles}, sample=sample
        )

    return new_dict


def filter_vcf_metrics(
    alleles: list[Allele],
    variant_metrics: dict[str, dict[str, str]],
    metric_name: str,
    metric_threshold: float,
    microarray: bool,
) -> tuple[defaultdict[str, list[Allele]], list[Allele]]:
    """Filter out alleles based on a specified read depth metric.

    Iterates through each allele's defining variants and compares the specified metric
    (e.g., "DP" for read depth) to a threshold value. For microarray data, the read depth
    is set to a constant value of 30.0. Alleles whose read depth falls below the threshold
    are collected in `filtered_out`; all others are placed in `passed_filtering`.

    Args:
        alleles (list[Allele]):
            A list of allele objects to be evaluated.
        variant_metrics (dict[str, dict[str, str]]):
            A nested dictionary where the key is a variant identifier and the value
            is a dictionary of metrics (e.g., {"DP": "45", ...}).
        metric_name (str):
            The name of the metric to evaluate (e.g., "DP" for read depth).
        metric_threshold (float):
            The threshold value for the chosen metric. Alleles below this value
            are excluded.
        microarray (bool):
            If True, overrides the chosen metric by setting read depth to 30.0.

    Returns:
        tuple[defaultdict[str, list[Allele]], list[Allele]]:
            A tuple containing two elements:
            1. `filtered_out`: A defaultdict where each key is
               "variant:read_depth" and each value is a list of alleles
               that did not meet the threshold.
            2. `passed_filtering`: A list of alleles that passed the threshold.

    Raises:
        KeyError:
            If a required variant or metric is missing in `variant_metrics`.
    """
    # TODO large dels will have depth zero
    filtered_out = defaultdict(list)
    passed_filtering = []
    metric_threshold = float(metric_threshold)
    for allele in alleles:
        keep = True
        for variant in allele.defining_variants:
            read_depth = float(variant_metrics[variant][metric_name])
            if microarray:
                read_depth = 30.0  # for microarray
            else:
                read_depth = float(variant_metrics[variant][metric_name])
            if read_depth < metric_threshold:
                filtered_out[f"{variant}:{str(read_depth)}"].append(allele)
                keep = False
        if keep:
            passed_filtering.append(allele)

    return filtered_out, passed_filtering


@apply_to_dict_values
def remove_alleles_with_low_read_depth(
    bg: BloodGroup,
    variant_metrics: dict[str, str],
    min_read_depth: int,
    microarray: bool,
) -> BloodGroup:
    """
    Remove alleles from a BloodGroup object that have defining variants with read depth
    below a specified minimum threshold.

    Args:
        bg (BloodGroup): The BloodGroup object containing alleles to filter.
        variant_metrics (dict[str, dict[str, int]]): A dictionary containing variant
        metrics with read depth information.
        min_read_depth (int): The minimum read depth threshold.

    Returns:
        BloodGroup: The BloodGroup object with alleles filtered based on read depth.
    """

    filtered_out, passed_filtering = filter_vcf_metrics(
        bg.alleles[AlleleState.RAW], variant_metrics, "DP", min_read_depth, microarray
    )
    if filtered_out:
        vars_affected = ",".join(filtered_out.keys())
        message = f"Read Depth. Sample: {bg.sample}, BG: {bg.type}, variant/s: {vars_affected}"
        logger.warning(message)
    bg.filtered_out["insufficient_read_depth"] = filtered_out
    bg.alleles[AlleleState.FILT] = passed_filtering
    return bg


@apply_to_dict_values
def remove_alleles_with_low_base_quality(
    bg: BloodGroup,
    variant_metrics: dict[str, str],
    min_base_quality: int,
    microarray: bool,
) -> BloodGroup:
    """
    Remove alleles from a BloodGroup object that have defining variants with base
    quality below a specified minimum threshold.

    Args:
        bg (BloodGroup): The BloodGroup object containing alleles to filter.
        variant_metrics (dict[str, dict[str, int]]): A dictionary containing variant
        metrics with read depth information.
        min_base_quality (int): The minimum base_quality threshold.

    Returns:
        BloodGroup: The BloodGroup object with alleles filtered based on read depth.
    """

    filtered_out, passed_filtering = filter_vcf_metrics(
        bg.alleles[AlleleState.FILT],
        variant_metrics,
        "GQ",
        min_base_quality,
        microarray,
    )
    if filtered_out:
        vars_affected = ",".join(filtered_out.keys())
        message = f"Base Quality. Sample: {bg.sample}, BG: {bg.type}, variant/s: {vars_affected}"
        logger.warning(message)
    bg.filtered_out["insufficient_min_base_quality"] = filtered_out
    bg.alleles[AlleleState.FILT] = passed_filtering

    return bg


def get_fully_homozygous_alleles(
    ranked_chunks: list[list[Allele]], variant_pool: dict[str, Any]
) -> list[list[Allele]]:
    """Filter out alleles that are not fully homozygous from a list of ranked allele chunks.

    Uses a partial function to check each allele's variants in the provided `variant_pool`.
    Only alleles where every relevant variant equals the required homozygous genotype (2)
    are included in the result.

    Args:
        ranked_chunks (list[list[Allele]]):
            A list of lists (chunks), where each chunk contains ranked Allele objects.
        variant_pool (dict[str, Any]):
            A dictionary containing variant data used for assessing homozygosity.
            The exact structure depends on the `check_available_variants` function.

    Returns:
        list[list[Allele]]:
            A list of lists, each mirroring the structure of `ranked_chunks`
            but including only alleles that are fully homozygous in every variant.

    Raises:
        KeyError:
            If a variant key is missing in `variant_pool`.
    """
    check_hom = partial(check_available_variants, 2, variant_pool, operator.eq)
    homs = [[] for _ in ranked_chunks]

    for i, chunk in enumerate(ranked_chunks):
        for allele in chunk:
            if all(check_hom(allele)):
                homs[i].append(allele)
    return homs


def unique_in_order(lst: list) -> list:
    """
    Return a list of unique elements from 'lst' in the order they first appear,
    without using a set or other unordered data structure.

    Args:
        lst: The input list (possibly with duplicates).

    Returns:
        A list of items from 'lst' with duplicates removed in order.

    Example:
        >>> unique_in_order([3, 3, 1, 2, 1, 3])
        [3, 1, 2]
    """
    unique_items = []
    for item in lst:
        # Append item only if it's not already in the unique list
        if item not in unique_items:
            unique_items.append(item)
    return unique_items


# -----------------------------------------------------------
# Protocol for structural subtyping
# -----------------------------------------------------------
class GeneticProcessingProtocol(Protocol):
    """Protocol defining a process method for genetic data."""

    def process(
        self, bg: BloodGroup, reference_alleles: dict[str, Allele]
    ) -> list[Pair]:
        """Process a BloodGroup and return `AlleleState.NORMAL` pairs."""
        ...


# -----------------------------------------------------------
# Concrete strategies
# -----------------------------------------------------------
@dataclass
class NoVariantStrategy:
    """Handles the case where there are no variants."""

    def process(
        self, bg: BloodGroup, reference_alleles: dict[str, Allele]
    ) -> list[Pair]:
        ref_allele = reference_alleles[bg.type]
        return [Pair(ref_allele, ref_allele)]


@dataclass
class SingleVariantStrategy:
    """Handles the case where there is a single variant."""

    def process(
        self, bg: BloodGroup, reference_alleles: dict[str, Allele]
    ) -> list[Pair]:
        return [
            make_pair(
                reference_alleles, bg.variant_pool_numeric, bg.alleles[AlleleState.FILT]
            )
        ]


@dataclass
class MultipleVariantDispatcher:
    """Chooses a sub-strategy when multiple variants are present."""

    def process(
        self, bg: BloodGroup, reference_alleles: dict[str, Allele]
    ) -> list[Pair]:
        options = unique_in_order(bg.alleles[AlleleState.FILT])
        non_ref_options = get_non_refs(options)
        ranked_chunks = chunk_geno_list_by_rank(non_ref_options)
        homs = get_fully_homozygous_alleles(ranked_chunks, bg.variant_pool_numeric)

        first_chunk = ranked_chunks[0]
        weight_first_chunk = first_chunk[0].weight_geno
        trumpiest_homs = homs[0]
        weight_trumpiest_homs = trumpiest_homs[0].weight_geno if trumpiest_homs else 1000
        
        # Sub-strategy selection
        if len(trumpiest_homs) == 1:
            return SingleHomMultiVariantStrategy(
                hom_allele=trumpiest_homs[0], first_chunk=first_chunk
            ).process(bg, reference_alleles)
        elif len(trumpiest_homs) > 1 and weight_first_chunk == weight_trumpiest_homs:
            return MultipleHomMultiVariantStrategy(
                homs=trumpiest_homs, first_chunk=first_chunk
            ).process(bg, reference_alleles)
        elif any(len(hom_chunk) > 0 for hom_chunk in homs):
            return SomeHomMultiVariantStrategy(ranked_chunks=ranked_chunks).process(
                bg, reference_alleles
            )
        else:
            return NoHomMultiVariantStrategy(non_ref_options=non_ref_options).process(
                bg, reference_alleles
            )


@dataclass
class SingleHomMultiVariantStrategy:
    hom_allele: Allele
    first_chunk: list[Allele]

    def process(
        self, bg: BloodGroup, reference_alleles: dict[str, Allele]
    ) -> list[Pair]:
        hom_pair = [Pair(self.hom_allele, self.hom_allele)]
        if len(self.first_chunk) == 1:
            return hom_pair
        elif any(self.hom_allele in allele for allele in self.first_chunk):
            return combine_all(self.first_chunk, bg.variant_pool_numeric)
        else:
            return hom_pair + combine_all(self.first_chunk, bg.variant_pool_numeric)


@dataclass
class MultipleHomMultiVariantStrategy:
    homs: list[Allele]
    first_chunk: list[Allele]

    def process(
        self, bg: BloodGroup, reference_alleles: dict[str, Allele]
    ) -> list[Pair]:
        new_pairs = [Pair(h, h) for h in self.homs]
        if len(self.first_chunk) > len(self.homs):
            return new_pairs + combine_all(self.first_chunk, bg.variant_pool_numeric)
        return new_pairs


@dataclass
class SomeHomMultiVariantStrategy:
    ranked_chunks: list[list[Allele]]
    def process(
        self, bg: BloodGroup, reference_alleles: dict[str, Allele]
    ) -> list[Pair]:
        homs = get_fully_homozygous_alleles(self.ranked_chunks, bg.variant_pool_numeric)
        if len(homs) > 2 and len(homs[0]) == 0 and len(homs[1]) == 0:
            raise ValueError("No homs in the first two rank tiers.")

        first_chunk = self.ranked_chunks[0]
        if len(first_chunk) == 1 and len(self.ranked_chunks) == 1:
            return [
                make_pair(
                    reference_alleles,
                    bg.variant_pool_numeric.copy(),
                    first_chunk,
                )
            ]
        return combine_all(
            self.ranked_chunks[0] + self.ranked_chunks[1], bg.variant_pool_numeric
        )


@dataclass
class NoHomMultiVariantStrategy:
    non_ref_options: list[Allele]

    def process(
        self, bg: BloodGroup, reference_alleles: dict[str, Allele]
    ) -> list[Pair]:
        ref_allele = reference_alleles[
            bg.type
        ]  # TODO - this is putting refs back in that have already been discarded as impossible
        ref_options = self.non_ref_options + [ref_allele]
        return combine_all(ref_options, bg.variant_pool_numeric)


# -----------------------------------------------------------
# Picking the right protocol-based strategy
# -----------------------------------------------------------
def _pick_strategy(bg: BloodGroup) -> GeneticProcessingProtocol:
    """Decide which strategy (protocol implementer) to use."""
    options = unique_in_order(bg.alleles[AlleleState.FILT])
    if len(options) == 0:
        return NoVariantStrategy()
    elif len(options) == 1:
        return SingleVariantStrategy()
    else:
        return MultipleVariantDispatcher()


@apply_to_dict_values
def process_genetic_data(
    bg: BloodGroup, reference_alleles: dict[str, Allele]
) -> BloodGroup:
    """Process genetic data to identify alleles and genotypes.

    Args:
        bg (BloodGroup):
            The blood group data that contains alleles (POS, NORMAL, etc.).
        reference_alleles (dict[str, Allele]):
            A dictionary mapping blood group types to reference Allele objects.

    Returns:
        An updated BloodGroup with `AlleleState.NORMAL` pairs set appropriately.

    Raises:
        ValueError: When constraints in the multiple-variant scenario are violated.
    """

    strategy: GeneticProcessingProtocol = _pick_strategy(
        bg
    )  # Returns a Protocol implementer
    normal_pairs = strategy.process(bg, reference_alleles)
    bg.alleles[AlleleState.NORMAL] = normal_pairs
    
    return bg


@apply_to_dict_values
def find_what_was_excluded_due_to_rank(
    bg: BloodGroup, reference_alleles: dict[str, Allele]
) -> BloodGroup:
    """Find all possible allele pairs based on genetic data.

    If the pairs are not present, list them in
    filtered_out["excluded_due_to_rank*"].

    Args:
        bg (BloodGroup): A BloodGroup object containing alleles, the variant pool,
            and other genetic data.
        reference_alleles (dict[str, Allele]): A dictionary mapping blood group types
            to their reference Allele.

    Returns:
        BloodGroup: The updated BloodGroup with pairs excluded due to rank added to
            the filtered_out collections.
    """

    options = set(bg.alleles[AlleleState.FILT])
    non_ref_options = get_non_refs(options)
    if non_ref_options:
        for pair in combine_all(non_ref_options, bg.variant_pool_numeric):
            if pair not in bg.alleles[AlleleState.NORMAL]:
                bg.filtered_out["excluded_due_to_rank"].append(pair)
        ref_options = non_ref_options + [
            reference_alleles[non_ref_options[0].blood_group]
        ]
        for pair in combine_all(ref_options, bg.variant_pool_numeric):
            if pair not in bg.alleles[AlleleState.NORMAL]:
                bg.filtered_out["excluded_due_to_rank_ref"].append(pair)
        ranked_chunks = chunk_geno_list_by_rank(non_ref_options)
        homs = get_fully_homozygous_alleles(ranked_chunks, bg.variant_pool_numeric)
        for ranked_homs in homs:
            for hom in ranked_homs:
                pair = Pair(allele1=hom, allele2=hom)
                if pair not in bg.alleles[AlleleState.NORMAL]:
                    bg.filtered_out["excluded_due_to_rank_hom"].append(pair)

    return bg


def make_pair(
    reference_alleles: dict[str, str], variant_pool: list[str], sub_results: list[str]
) -> list[str]:
    """Creates a pair of alleles based on the given parameters.

    Args:
        reference_alleles (Dict[str, str]): A mapping from blood group to reference
        allele.
        variant_pool (List[str]): A list of available variants.
        sub_results (List[str]): A list containing the initial results, expected to be
        of length 1.

    Returns:
        List[str]: A list containing the original results and an additional allele,
        either a duplicate of the first (if checks pass) or a corresponding
        reference allele.

    Raises:
        AssertionError: If the length of `sub_results` is not 1.
    """
    sub_results = list(sub_results)
    check_vars = partial(check_available_variants, 2, variant_pool, operator.eq)
    assert len(sub_results) == 1
    if all(check_vars(sub_results[0])):  # this is essentially fully_hom (func)
        sub_results.append(sub_results[0])
    else:
        sub_results.append(reference_alleles[sub_results[0].blood_group])
    return Pair(*sub_results)


def pair_can_exist(
    pair: tuple[Allele, Allele], variant_pool_copy: dict[str, int]
) -> bool:
    """Check if a pair of alleles can exist based on the variant pool.

    NB: This is a bit of a misnomer, as it only subtracts in more complex cases,
    like "009Kenya A4GALT": A4GALT*01/A4GALT*02 is not possible because if
    'A4GALT*02' then 22:43089849_T_C is on the other side so it has to be
    'A4GALT*01.02' and not reference.

    Args:
        pair (tuple[Allele, Allele]): A tuple containing two Allele objects.
        variant_pool_copy (dict[str, int]): A dictionary mapping variant identifiers
            to their available counts.

    Returns:
        bool: True if the pair can exist based on the variant pool, False otherwise.
    """
    allele1, allele2 = pair
    if allele1.reference or allele2.reference:
        return True
    for variant in allele1.defining_variants:
        variant_pool_copy[variant] -= 1
    return all(variant_pool_copy[variant] >= 1 for variant in allele2.defining_variants)


def combine_all(alleles: list[Allele], variant_pool: dict[str, int]) -> list[Pair]:
    """Combine all alleles into pairs, if possible.

    Args:
        alleles (list[Allele]): A list of Allele objects to be paired.
        variant_pool (dict[str, int]): A dictionary mapping variant identifiers
            to their available counts.

    Returns:
        list[Pair]: A list of Pair objects where each pair satisfies the variant
            pool constraints.
    """
    ranked = []
    for pair in itertools.combinations(alleles, 2):
        if pair_can_exist(pair, variant_pool.copy()):
            ranked.append(Pair(*pair))
    return ranked


@apply_to_dict_values
def add_CD_to_XG(bg: BloodGroup) -> BloodGroup:
    """
    adds CD to XG blood group.

    Args:
        bg (BloodGroup): The BloodGroup object to be processed.

    Returns:
        BloodGroup: The processed BloodGroup object.
    """
    if bg.genotypes == ["XG*01/XG*01"]:
        bg.genotypes = ["XG*01/XG*01", "CD99*01/CD99*01"]
    return bg


def add_refs(db: Db, res: dict[str, BloodGroup], exclude) -> dict[str, BloodGroup]:
    """Add reference genotypes to existing results or create new entries for them.

    Args:
        db (Db): The database object containing reference alleles.
        res (Dict[str, BloodGroup]): Dictionary of BloodGroup objects to be updated
        with reference data.

    Returns:
        Dict[str, BloodGroup]: The updated dictionary of BloodGroup objects with added
          reference genotypes.

    This function checks for existing blood groups in the results dictionary and adds
    the reference genotype from the database if not present. It initializes a new
    BloodGroup object for any blood group type not already included in the results with
    the reference genotype as both a 'raw' and 'paired' allele.
    """
    for blood_group, reference in db.reference_alleles.items():
        if blood_group in exclude:
            continue
        if blood_group not in res:
            res[blood_group] = BloodGroup(
                type=blood_group,
                alleles={
                    AlleleState.RAW: [reference],
                    AlleleState.FILT: [reference],
                    AlleleState.NORMAL: [Pair(*[reference] * 2)],
                },
                sample="ref",
                genotypes=[f"{reference.genotype}/{reference.genotype}"],
            )
    return res



#!/usr/bin/env python3
"""
Optimized VCF Validator Module with Faster Control Character Check
------------------------------
This module validates VCF files for correct formatting and scans for malicious content in a single pass.
It replaces the regex-based control character check with a set-based lookup, which is faster.
"""

import os
import gzip
from dataclasses import dataclass

# Precompute a set of disallowed control characters (allowed: tab, newline, carriage return)
SUSPICIOUS_CHARS = {chr(c) for c in list(range(0, 9)) + list(range(11, 13)) + list(range(14, 32))}


@dataclass
class VCFValidationResult:
    """
    A data class representing the result of VCF validation.
    ----------
    Attributes
    ----------
    is_valid (bool):
        Indicates if the validation passed.
    errors (list[str]):
        List of error messages encountered during validation.
    """
    is_valid: bool
    errors: list[str]


def open_vcf_file(file_path: str, mode: str = "rt"):
    """
    Open a VCF file that may be plain text or gzip compressed.
    
    Args:
        file_path (str): Path to the VCF file.
        mode (str, optional): File open mode. Defaults to "rt".
    
    Returns:
        A file object.
    """
    if str(file_path).endswith(".gz"):
        return gzip.open(file_path, mode, encoding="utf-8")
    return open(file_path, mode, encoding="utf-8")


class VCFValidator:
    """
    A class for validating VCF files in a single pass.
    """
    MAX_FILE_SIZE = 2 * 1024 * 1024 * 1024  # 2GB file size limit

    @staticmethod
    def validate_file(file_path: str) -> VCFValidationResult:
        """
        Validate the VCF file for correct format and malicious content in a single pass.
        
        Args:
            file_path (str): Path to the VCF file.
        
        Returns:
            VCFValidationResult: Result of validation.
        """
        errors: list[str] = []

        if not os.path.exists(file_path):
            return VCFValidationResult(False, ["File does not exist."])

        file_size: int = os.path.getsize(file_path)
        if file_size > VCFValidator.MAX_FILE_SIZE:
            return VCFValidationResult(False, ["File size exceeds allowed limit."])

        found_fileformat = False
        header_line: str | None = None

        try:
            with open_vcf_file(file_path, "rt") as f:
                for line in f:
                    # Check for malicious content.
                    if "\0" in line:
                        errors.append("File contains null byte(s).")
                    # Replace regex search with set intersection for control characters.
                    if set(line) & SUSPICIOUS_CHARS:
                        errors.append("File contains suspicious control characters.")
                    if "<script" in line.lower():
                        errors.append("File contains embedded script tags.")

                    # Check for meta-information header.
                    if not found_fileformat and line.startswith("##fileformat=VCF"):
                        found_fileformat = True
                    # Find header line: first line starting with '#' but not '##'.
                    if header_line is None and line.startswith("#") and not line.startswith("##"):
                        header_line = line.strip()
        except UnicodeDecodeError:
            return VCFValidationResult(False, ["File is not valid UTF-8 text."])
        except Exception as e:
            return VCFValidationResult(False, [f"Error reading file: {str(e)}"])

        if not found_fileformat:
            errors.append("Missing '##fileformat=VCF' header.")

        if header_line is None:
            errors.append("Missing header line starting with '#CHROM'.")
        else:
            columns = header_line.split("\t")
            if len(columns) < 8:
                errors.append("Header line has insufficient columns.")
            else:
                required = ["#CHROM", "POS", "ID", "REF", "ALT", "QUAL", "FILTER", "INFO"]
                for i, col in enumerate(required):
                    if columns[i] != col:
                        errors.append(f"Header column {i + 1} expected '{col}', found '{columns[i]}'.")
                        break

        # Remove duplicate errors.
        unique_errors = list(dict.fromkeys(errors))
        return VCFValidationResult(len(unique_errors) == 0, unique_errors)


def validate_vcf(file_path: str) -> VCFValidationResult:
    """
    Validate a VCF file for proper format and absence of malicious content.
    
    Args:
        file_path (str): Path to the VCF file.
    
    Returns:
        VCFValidationResult: Combined result of format and security checks.
    """
    return VCFValidator.validate_file(file_path)